@@ -1,23 +1,16 @@
 # This file is dual licensed under the terms of the Apache License, Version
 # 2.0, and the BSD License. See the LICENSE file in the root of this repository
 # for complete details.
-from __future__ import absolute_import, division, print_function
 
 import itertools
 import operator
-import warnings
 
 import pytest
 
-from packaging.specifiers import (
-    InvalidSpecifier,
-    LegacySpecifier,
-    Specifier,
-    SpecifierSet,
-)
-from packaging.version import LegacyVersion, Version, parse
+from packaging.specifiers import InvalidSpecifier, Specifier, SpecifierSet
+from packaging.version import Version, parse
 
-from .test_version import LEGACY_VERSIONS, VERSIONS
+from .test_version import VERSIONS
 
 LEGACY_SPECIFIERS = [
     "==2.1.0.3",
@@ -38,7 +31,6 @@ SPECIFIERS = [
     ">=7.9a1",
     "<1.0.dev1",
     ">2.0.post1",
-    "===lolwat",
 ]
 
 
@@ -72,7 +64,14 @@ class TestSpecifier:
             # support one or the other
             "==1.0.*+5",
             "!=1.0.*+deadbeef",
-            # Prefix matching cannot be used inside of a local version
+            # Prefix matching cannot be used with a pre-release, post-release,
+            # dev or local version
+            "==2.0a1.*",
+            "!=2.0a1.*",
+            "==2.0.post1.*",
+            "!=2.0.post1.*",
+            "==2.0.dev1.*",
+            "!=2.0.dev1.*",
             "==1.0+5.*",
             "!=1.0+deadbeef.*",
             # Prefix matching must appear at the end
@@ -228,7 +227,7 @@ class TestSpecifier:
         spec = Specifier(specifier)
 
         assert str(spec) == expected
-        assert repr(spec) == "<Specifier({0})>".format(repr(expected))
+        assert repr(spec) == f"<Specifier({expected!r})>"
 
     @pytest.mark.parametrize("specifier", SPECIFIERS)
     def test_specifiers_hash(self, specifier):
@@ -308,12 +307,15 @@ class TestSpecifier:
                 ("2b1.dev1", "==2.*"),
                 ("2c1", "==2.*"),
                 ("2c1.post1.dev1", "==2.*"),
+                ("2c1.post1.dev1", "==2.0.*"),
                 ("2rc1", "==2.*"),
+                ("2rc1", "==2.0.*"),
                 ("2", "==2.*"),
+                ("2", "==2.0.*"),
+                ("2", "==0!2.*"),
+                ("0!2", "==2.*"),
                 ("2.0", "==2.*"),
                 ("2.0.0", "==2.*"),
-                ("2.0.post1", "==2.0.post1.*"),
-                ("2.0.post1.dev1", "==2.0.post1.*"),
                 ("2.1+local.version", "==2.1.*"),
                 # Test the in-equality operation
                 ("2.1", "!=2"),
@@ -360,6 +362,8 @@ class TestSpecifier:
                 ("1.0.1", "~=1.0"),
                 ("1.1", "~=1.0"),
                 ("1.9999999", "~=1.0"),
+                ("1.1", "~=1.0a1"),
+                ("2022.01.01", "~=2022.01.01"),
                 # Test that epochs are handled sanely
                 ("2!1.0", "~=2!1.0"),
                 ("2!1.0", "==2!1.*"),
@@ -404,12 +408,13 @@ class TestSpecifier:
                 ("2b1.dev1", "!=2.*"),
                 ("2c1", "!=2.*"),
                 ("2c1.post1.dev1", "!=2.*"),
+                ("2c1.post1.dev1", "!=2.0.*"),
                 ("2rc1", "!=2.*"),
+                ("2rc1", "!=2.0.*"),
                 ("2", "!=2.*"),
+                ("2", "!=2.0.*"),
                 ("2.0", "!=2.*"),
                 ("2.0.0", "!=2.*"),
-                ("2.0.post1", "!=2.0.post1.*"),
-                ("2.0.post1.dev1", "!=2.0.post1.*"),
                 # Test the greater than equal operation
                 ("2.0.dev1", ">=2"),
                 ("2.0a1", ">=2"),
@@ -489,12 +494,10 @@ class TestSpecifier:
     @pytest.mark.parametrize(
         ("version", "spec", "expected"),
         [
-            # Test identity comparison by itself
-            ("lolwat", "===lolwat", True),
-            ("Lolwat", "===lolwat", True),
-            ("1.0", "===1.0", True),
-            ("nope", "===lolwat", False),
             ("1.0.0", "===1.0", False),
+            ("1.0.dev0", "===1.0", False),
+            # Test identity comparison by itself
+            ("1.0", "===1.0", True),
             ("1.0.dev0", "===1.0.dev0", True),
         ],
     )
@@ -535,7 +538,6 @@ class TestSpecifier:
             (">=1.0", "2.0.dev1", False),
             (">=2.0.dev1", "2.0a1", True),
             ("==2.0.*", "2.0a1.dev1", False),
-            ("==2.0a1.*", "2.0a1.dev1", True),
             ("<=2.0", "1.0.dev1", False),
             ("<=2.0.dev1", "1.0a1", True),
         ],
@@ -567,10 +569,6 @@ class TestSpecifier:
 
         assert list(spec.filter(input, **kwargs)) == expected
 
-    @pytest.mark.xfail
-    def test_specifier_explicit_legacy(self):
-        assert Specifier("==1.0").contains(LegacyVersion("1.0"))
-
     @pytest.mark.parametrize(
         ("spec", "op"),
         [
@@ -583,6 +581,7 @@ class TestSpecifier:
             (">=7.9a1", ">="),
             ("<1.0.dev1", "<"),
             (">2.0.post1", ">"),
+            # === is an escape hatch in PEP 440
             ("===lolwat", "==="),
         ],
     )
@@ -601,6 +600,7 @@ class TestSpecifier:
             (">=7.9a1", "7.9a1"),
             ("<1.0.dev1", "1.0.dev1"),
             (">2.0.post1", "2.0.post1"),
+            # === is an escape hatch in PEP 440
             ("===lolwat", "lolwat"),
         ],
     )
@@ -627,145 +627,18 @@ class TestSpecifier:
     )
     def test_iteration(self, spec, expected_items):
         spec = SpecifierSet(spec)
-        items = set(str(item) for item in spec)
+        items = {str(item) for item in spec}
         assert items == set(expected_items)
 
+    def test_specifier_equal_for_compatible_operator(self):
+        assert Specifier("~=1.18.0") != Specifier("~=1.18")
 
-class TestLegacySpecifier:
-    def test_legacy_specifier_is_deprecated(self):
-        with warnings.catch_warnings(record=True) as w:
-            LegacySpecifier(">=some-legacy-version")
-            assert len(w) == 1
-            assert issubclass(w[0].category, DeprecationWarning)
-
-    @pytest.mark.parametrize(
-        ("version", "spec", "expected"),
-        [
-            (v, s, True)
-            for v, s in [
-                # Test the equality operation
-                ("2.0", "==2"),
-                ("2.0", "==2.0"),
-                ("2.0", "==2.0.0"),
-                # Test the in-equality operation
-                ("2.1", "!=2"),
-                ("2.1", "!=2.0"),
-                ("2.0.1", "!=2"),
-                ("2.0.1", "!=2.0"),
-                ("2.0.1", "!=2.0.0"),
-                # Test the greater than equal operation
-                ("2.0", ">=2"),
-                ("2.0", ">=2.0"),
-                ("2.0", ">=2.0.0"),
-                ("2.0.post1", ">=2"),
-                ("2.0.post1.dev1", ">=2"),
-                ("3", ">=2"),
-                # Test the less than equal operation
-                ("2.0", "<=2"),
-                ("2.0", "<=2.0"),
-                ("2.0", "<=2.0.0"),
-                ("2.0.dev1", "<=2"),
-                ("2.0a1", "<=2"),
-                ("2.0a1.dev1", "<=2"),
-                ("2.0b1", "<=2"),
-                ("2.0b1.post1", "<=2"),
-                ("2.0c1", "<=2"),
-                ("2.0c1.post1.dev1", "<=2"),
-                ("2.0rc1", "<=2"),
-                ("1", "<=2"),
-                # Test the greater than operation
-                ("3", ">2"),
-                ("2.1", ">2.0"),
-                # Test the less than operation
-                ("1", "<2"),
-                ("2.0", "<2.1"),
-            ]
-        ]
-        + [
-            (v, s, False)
-            for v, s in [
-                # Test the equality operation
-                ("2.1", "==2"),
-                ("2.1", "==2.0"),
-                ("2.1", "==2.0.0"),
-                # Test the in-equality operation
-                ("2.0", "!=2"),
-                ("2.0", "!=2.0"),
-                ("2.0", "!=2.0.0"),
-                # Test the greater than equal operation
-                ("2.0.dev1", ">=2"),
-                ("2.0a1", ">=2"),
-                ("2.0a1.dev1", ">=2"),
-                ("2.0b1", ">=2"),
-                ("2.0b1.post1", ">=2"),
-                ("2.0c1", ">=2"),
-                ("2.0c1.post1.dev1", ">=2"),
-                ("2.0rc1", ">=2"),
-                ("1", ">=2"),
-                # Test the less than equal operation
-                ("2.0.post1", "<=2"),
-                ("2.0.post1.dev1", "<=2"),
-                ("3", "<=2"),
-                # Test the greater than operation
-                ("1", ">2"),
-                ("2.0.dev1", ">2"),
-                ("2.0a1", ">2"),
-                ("2.0a1.post1", ">2"),
-                ("2.0b1", ">2"),
-                ("2.0b1.dev1", ">2"),
-                ("2.0c1", ">2"),
-                ("2.0c1.post1.dev1", ">2"),
-                ("2.0rc1", ">2"),
-                ("2.0", ">2"),
-                # Test the less than operation
-                ("3", "<2"),
-            ]
-        ],
-    )
-    def test_specifiers(self, version, spec, expected):
-        spec = LegacySpecifier(spec, prereleases=True)
-
-        if expected:
-            # Test that the plain string form works
-            assert version in spec
-            assert spec.contains(version)
-
-            # Test that the version instance form works
-            assert LegacyVersion(version) in spec
-            assert spec.contains(LegacyVersion(version))
-        else:
-            # Test that the plain string form works
-            assert version not in spec
-            assert not spec.contains(version)
-
-            # Test that the version instance form works
-            assert LegacyVersion(version) not in spec
-            assert not spec.contains(LegacyVersion(version))
-
-    def test_specifier_explicit_prereleases(self):
-        spec = LegacySpecifier(">=1.0")
-        assert not spec.prereleases
-        spec.prereleases = True
-        assert spec.prereleases
-
-        spec = LegacySpecifier(">=1.0", prereleases=False)
-        assert not spec.prereleases
-        spec.prereleases = True
-        assert spec.prereleases
-
-        spec = LegacySpecifier(">=1.0", prereleases=True)
-        assert spec.prereleases
-        spec.prereleases = False
-        assert not spec.prereleases
-
-        spec = LegacySpecifier(">=1.0", prereleases=True)
-        assert spec.prereleases
-        spec.prereleases = None
-        assert not spec.prereleases
+    def test_specifier_hash_for_compatible_operator(self):
+        assert hash(Specifier("~=1.18.0")) != hash(Specifier("~=1.18"))
 
 
 class TestSpecifierSet:
-    @pytest.mark.parametrize("version", VERSIONS + LEGACY_VERSIONS)
+    @pytest.mark.parametrize("version", VERSIONS)
     def test_empty_specifier(self, version):
         spec = SpecifierSet(prereleases=True)
 
@@ -813,6 +686,15 @@ class TestSpecifierSet:
         assert spec.contains("1.0.dev1")
         assert not spec.contains("1.0.dev1", prereleases=False)
 
+    def test_specifier_contains_installed_prereleases(self):
+        spec = SpecifierSet("~=1.0")
+        assert not spec.contains("1.0.0.dev1", installed=True)
+        assert spec.contains("1.0.0.dev1", prereleases=True, installed=True)
+
+        spec = SpecifierSet("~=1.0", prereleases=True)
+        assert spec.contains("1.0.0.dev1", installed=True)
+        assert not spec.contains("1.0.0.dev1", prereleases=False, installed=False)
+
     @pytest.mark.parametrize(
         ("specifier", "specifier_prereleases", "prereleases", "input", "expected"),
         [
@@ -821,7 +703,6 @@ class TestSpecifierSet:
             (">=1.0.dev1", None, None, ["1.0", "2.0a1"], ["1.0", "2.0a1"]),
             ("", None, None, ["1.0a1"], ["1.0a1"]),
             ("", None, None, ["1.0", Version("2.0")], ["1.0", Version("2.0")]),
-            ("", None, None, ["2.0dog", "1.0"], ["1.0"]),
             # Test overriding with the prereleases parameter on filter
             ("", None, False, ["1.0a1"], []),
             (">=1.0.dev1", None, False, ["1.0", "2.0a1"], ["1.0"]),
@@ -847,10 +728,6 @@ class TestSpecifierSet:
 
         assert list(spec.filter(input, **kwargs)) == expected
 
-    def test_legacy_specifiers_combined(self):
-        spec = SpecifierSet("<3,>1-1-1")
-        assert "2.0" in spec
-
     @pytest.mark.parametrize(
         ("specifier", "expected"),
         [
@@ -873,7 +750,7 @@ class TestSpecifierSet:
         spec = SpecifierSet(specifier)
 
         assert str(spec) == expected
-        assert repr(spec) == "<SpecifierSet({0})>".format(repr(expected))
+        assert repr(spec) == f"<SpecifierSet({expected!r})>"
 
     @pytest.mark.parametrize("specifier", SPECIFIERS + LEGACY_SPECIFIERS)
     def test_specifiers_hash(self, specifier):
@@ -996,3 +873,8 @@ class TestSpecifierSet:
     )
     def test_comparison_ignores_local(self, version, specifier, expected):
         assert (Version(version) in SpecifierSet(specifier)) == expected
+
+    def test_contains_with_compatible_operator(self):
+        combination = SpecifierSet("~=1.18.0") & SpecifierSet("~=1.18")
+        assert "1.19.5" not in combination
+        assert "1.18.0" in combination
